
import data.Graph;
import gnu.trove.set.hash.TCustomHashSet;
import io.FileManager;
import matching.RISolver;

public class MultiRI
{
    public static void main(String[] args)
    {
        //Input parameters
        String netFile=null;
        String queryFile=null;
        String dumpFile=null;

        int i;
        //Read input parameters
        for (i=0;i<args.length;i++)
        {
            switch (args[i])
            {
                case "-n": netFile=args[++i]; break;
                case "-q": queryFile=args[++i]; break;
                case "-d": dumpFile=args[++i]; break;
                default: System.out.println("Error! Unrecognizable command '" + args[i] + "'"); printHelp(); System.exit(1);
            }
        }
        //Error in case network file is missing
        if(netFile==null)
        {
            System.out.println("Error! No input network has been specified!\n");
            printHelp();
            System.exit(1);
        }
        if(queryFile==null)
        {
            System.out.println("Error! No query has been specified!\n");
            printHelp();
            System.exit(1);
        }

        //Read input network
        System.out.println("-----------------------------");
        System.out.println("GRAPH FILE: "+netFile);
        FileManager fm=new FileManager();
        if(!netFile.endsWith("geu") && !netFile.endsWith("ged"))
        {
            System.out.println("Error! Unrecognizable file extension: "+netFile);
            System.exit(1);
        }
        Graph net=fm.readGraph(netFile);
        //System.out.println(fm.getSetLabels());
        //System.out.println(fm.getMapNodeLabels());
        //System.out.println(fm.getMapEdgeLabels());

        //Read queries
        System.out.println("QUERY FILE: "+queryFile);
        if(!queryFile.endsWith("geu") && !queryFile.endsWith("ged"))
        {
            System.out.println("Error! Unrecognizable file extension: "+queryFile);
            System.exit(1);
        }
        TCustomHashSet<Graph> setQueries=fm.readQueries(queryFile);

        //Perform subgraph matching
        System.out.println("Performing subgraph matching...");

        if(dumpFile!=null)
            fm.initDumpFile(dumpFile);
        for(Graph query : setQueries)
        {
            fm.printQuery(query);
            if(dumpFile!=null)
                fm.writeQuery(query);
            double inizio=System.currentTimeMillis();
            RISolver solver=new RISolver(net,dumpFile,fm);
            solver.match(query);
            double fine=System.currentTimeMillis();
            double totalTime=(fine-inizio)/1000;
            long numOccs=solver.getNumOccurrences();
            System.out.println("Number of matches found: "+numOccs);
            System.out.println("Time elapsed: "+totalTime+" secs");
        }
        if(dumpFile!=null)
            fm.closeDumpFile();

    }

    private static void printHelp()
    {
        String help = "Usage: java -jar MultiRI.jar -n <networkFile> -q <queryFile> [-d <embeddingsFile>]\n\n";
        help+="REQUIRED PARAMETERS:\n";
        help+="-n\tInput network file\n";
        help+="-q\tQueries file\n\n";
        help+="OPTIONAL PARAMETERS:\n";
        help+="-d\tSave embeddings found to output file\n";
        System.out.println(help);
    }
}
