package data;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.hash.TIntObjectHashMap;
import gnu.trove.set.hash.TIntHashSet;

import java.util.Arrays;
import java.util.Vector;

public class Graph
{
    private boolean directed;
    private TIntHashSet[] nodeLabels;
    private TIntObjectHashMap<TIntHashSet>[] outAdjList;
    private TIntObjectHashMap<TIntHashSet>[] inAdjList;
    private int multiplicity;

    public Graph(boolean directed, int numNodes)
    {
        this.directed=directed;
        this.nodeLabels=new TIntHashSet[numNodes];
        this.outAdjList=new TIntObjectHashMap[numNodes];
        this.inAdjList=new TIntObjectHashMap[numNodes];
        for(int i=0;i<numNodes;i++)
        {
            nodeLabels[i] = new TIntHashSet();
            outAdjList[i] = new TIntObjectHashMap<>();
            inAdjList[i] = new TIntObjectHashMap<>();
        }
        this.multiplicity=0;
    }

    public void addNode(int id, TIntHashSet label)
    {
        nodeLabels[id]=label;
    }

    public void addEdge(int source, int dest, TIntHashSet edgeLabels)
    {
        outAdjList[source].put(dest,edgeLabels);
        inAdjList[dest].put(source,edgeLabels);
        //System.out.println(source+"-"+dest+" --> "+setCols);
        if(!directed)
        {
            outAdjList[dest].put(source,edgeLabels);
            inAdjList[source].put(dest,edgeLabels);
        }
        if(edgeLabels.size()>multiplicity)
            multiplicity=edgeLabels.size();
    }

    public TIntHashSet[] getNodeLabels()
    {
        return nodeLabels;
    }

    public TIntObjectHashMap<TIntHashSet>[] getOutAdjList()
    {
        return outAdjList;
    }

    public TIntObjectHashMap<TIntHashSet>[] getInAdjList()
    {
        return inAdjList;
    }

    public int getNumNodes()
    {
        return outAdjList.length;
    }

    public boolean isDirected()
    {
        return directed;
    }

    public Vector<Integer>[] getSymmetryConditions()
    {
        int i, j, k;
        Vector<int[]> vv=findAutomorphisms();
        int numNodes=outAdjList.length;
        Vector<Integer>[] listCond=new Vector[numNodes];
        for(i=0;i<listCond.length;i++)
            listCond[i]=new Vector<>();
        int vvsize=vv.size();
        boolean[] broken=new boolean[vvsize];
        for(i=0;i<numNodes;i++)
        {
            for(j=0;j<vvsize;j++)
            {
                if(!broken[j] && vv.get(j)[i]!=i)
                    break;
            }
            if(j<vvsize)
            {
                for(k=i+1;k<numNodes;k++)
                {
                    for (j=0;j<vvsize;j++)
                    {
                        if(!broken[j] && vv.get(j)[i]==k)
                        {
                            listCond[k].add(i);
                            break;
                        }
                    }
                }
            }
            for(j=0;j<vvsize;j++)
            {
                if(vv.get(j)[i]!=i)
                    broken[j]=true;
            }
        }
        return listCond;
    }

    private Vector<int[]> findAutomorphisms()
    {
        int i, j, k;
        int g;
        int nof_nodes = nodeLabels.length;
        int[] fDir = new int[nof_nodes];
        int[] fRev = new int[nof_nodes];
        for (i = 0; i < nof_nodes; i++) {
            fDir[i] = -1;
            fRev[i] = -1;
        }
        int[][] sequence = new int[nof_nodes][nof_nodes];
        for (i = 0; i < nof_nodes; i++)
        {
            for (j = 0; j < nof_nodes; j++)
            {
                if (outAdjList[i].containsKey(j) || outAdjList[j].containsKey(i))
                {
                    int numNeighs = 0;
                    for (k = 0; k < nof_nodes; k++)
                    {
                        if (outAdjList[j].containsKey(k))
                            numNeighs++;
                        if (outAdjList[k].containsKey(j) && !outAdjList[j].containsKey(k))
                            numNeighs++;
                    }
                    sequence[i][j] = numNeighs;
                } else {
                    sequence[i][j] = 0;
                }
            }
        }
        for (i = 0; i < nof_nodes; i++)
            Arrays.sort(sequence[i]);

        boolean[] support = new boolean[nof_nodes * nof_nodes];
        for (i = 0; i < nof_nodes * nof_nodes; i++)
            support[i] = false;
        for (i = 0; i < nof_nodes; i++)
        {
            for (j = 0; j < nof_nodes; j++)
            {
                for (k = 0; k < nof_nodes; k++)
                {
                    if (sequence[i][k] != sequence[j][k])
                        break;
                }
                if (k >= nof_nodes)
                    support[i * nof_nodes + j] = true;
            }
        }

        Vector<int[]> vv=new Vector<>();
        for (g = 0; g < nof_nodes; g++)
        {
            if (support[g * nof_nodes])
            {
                fDir[0] = g;
                fRev[g] = 0;
                int pos = 1;
                isomorphicExtensions(fDir, fRev, vv, support, pos);
                fRev[fDir[0]] = -1;
                fDir[0] = -1;
            }
        }
        return vv;
    }

    private void isomorphicExtensions(int[] fDir, int[] fRev, Vector<int[]> vv, boolean[] support, int pos)
    {
        int i, j;
        int nof_nodes = fDir.length;
        int[] cand = new int[nof_nodes];
        int ncand;
        int num;
        for (i = 0; i < nof_nodes; i++)
            cand[i] = -1;
        if (pos == nof_nodes)
        {
            int[] vTemp=new int[nof_nodes];
            for(i=0;i<fDir.length;i++)
                vTemp[i]=fDir[i];
            vv.add(vTemp);
        }
        else
        {
            int n, m;
            boolean flag;
            int[] count = new int[nof_nodes];
            ncand = 0;
            for (i = 0; i < nof_nodes; i++)
                count[i] = 0;
            for (i = 0; i < nof_nodes; i++)
            {
                if (fDir[i] != -1)
                {
                    Vector<Integer> vNei = new Vector<>();
                    for (j = 0; j < nof_nodes; j++)
                    {
                        if (outAdjList[i].containsKey(j))
                            vNei.add(j);
                        else if (outAdjList[j].containsKey(i))
                            vNei.add(j);
                    }
                    num = vNei.size();
                    for (j = 0; j < num; j++)
                    {
                        int neigh = vNei.get(j);
                        if (fDir[neigh] == -1)
                        {
                            if (count[neigh] == 0)
                                cand[ncand++] = neigh;
                            count[neigh]++;
                        }
                    }
                }
            }
            m = 0;
            for (i = 1; i < ncand; i++)
            {
                if (count[i] > count[m])
                    m = i;
            }
            m = cand[m];
            ncand = 0;
            boolean[] already = new boolean[nof_nodes];
            for (i = 0; i < nof_nodes; i++)
                already[i] = false;
            for (i = 0; i < nof_nodes; i++)
            {
                if (fDir[i] != -1)
                {
                    Vector<Integer> vNei = new Vector<>();
                    for (j = 0; j < nof_nodes; j++)
                    {
                        if (outAdjList[fDir[i]].containsKey(j))
                            vNei.add(j);
                        else if (outAdjList[j].containsKey(fDir[i]))
                            vNei.add(j);
                    }
                    num = vNei.size();
                    for (j = 0; j < num; j++)
                    {
                        int neigh = vNei.get(j);
                        if (!already[neigh] && fRev[neigh] == -1 && support[m * nof_nodes + neigh])
                        {
                            cand[ncand++] = neigh;
                            already[neigh] = true;
                        }
                    }
                }
            }
            for (i = 0; i < ncand; i++)
            {
                n = cand[i];
                flag = false;
                if (nodeLabels[n].equals(nodeLabels[m]))
                {
                    for (j = 0; j < nof_nodes; j++)
                    {
                        if (fDir[j] != -1)
                        {
                            TIntHashSet edgeLabs1=outAdjList[m].get(j);
                            TIntHashSet edgeLabs2=outAdjList[n].get(fDir[j]);
                            if ((edgeLabs1==null && edgeLabs2!=null) ||
                                    (edgeLabs1!=null && edgeLabs2==null) ||
                                    (edgeLabs1!=null && !edgeLabs1.equals(edgeLabs2)))
                            {
                                flag = true;
                                break;
                            }
                            else
                            {
                                edgeLabs1=outAdjList[j].get(m);
                                edgeLabs2=outAdjList[fDir[j]].get(n);
                                if ((edgeLabs1==null && edgeLabs2!=null) ||
                                        (edgeLabs1!=null && edgeLabs2==null) ||
                                        (edgeLabs1!=null && !edgeLabs1.equals(edgeLabs2)))
                                {
                                    flag = true;
                                    break;
                                }
                            }

                        }
                    }
                    if (!flag)
                    {
                        fDir[m] = n;
                        fRev[n] = m;
                        pos++;
                        isomorphicExtensions(fDir, fRev, vv, support, pos);
                        pos--;
                        fRev[fDir[m]] = -1;
                        fDir[m] = -1;
                    }
                }
            }
        }
    }

    public String toString()
    {
        String str="";
        int i;
        for(i=0;i<outAdjList.length;i++)
        {
            str+="("+i+","+nodeLabels[i]+") --> ";
            TIntIterator it=outAdjList[i].keySet().iterator();
            while(it.hasNext())
            {
                int dest=it.next();
                str += "(" + dest;
                TIntHashSet setEdgeCols=outAdjList[i].get(dest);
                TIntIterator itCols=setEdgeCols.iterator();
                while(itCols.hasNext())
                {
                    int lab=itCols.next();
                    str+=","+lab;
                }
                str+=") ";
            }
            str+="\n";
        }
        return str;
    }

}
