package io;

import data.GraphStrategy;
import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.hash.TIntObjectHashMap;
import gnu.trove.set.hash.TCustomHashSet;
import gnu.trove.set.hash.TIntHashSet;

import java.io.*;
import java.util.HashMap;
import java.util.Vector;
import data.Graph;

public class FileManager
{
    private HashMap<String,Integer> mapNodeLabels;
    private HashMap<String,Integer> mapEdgeLabels;
    private Vector<String> setLabels;
    private BufferedWriter bwDump;
    private int maxMultiplicity;

    public FileManager()
    {
        this.mapNodeLabels=new HashMap<>();
        this.setLabels=new Vector<>();
        this.mapEdgeLabels=new HashMap<>();
        this.maxMultiplicity=0;
    }

    public Graph readGraph(String graphFile)
    {
        Graph g=null;
        try
        {
            BufferedReader br=new BufferedReader(new FileReader(graphFile));
            br.readLine();
            String str;
            boolean directed=false;
            if(graphFile.endsWith("ged"))
                directed=true;
            int numNodes=Integer.parseInt(br.readLine());
            g=new Graph(directed,numNodes);

            //Read node labels
            int i, j;
            for(i=0;i<numNodes;i++)
            {
                String[] split=br.readLine().split(",");
                TIntHashSet nodeLabs=new TIntHashSet();
                for(j=0;j<split.length;j++)
                {
                    String lab=split[j];
                    if(!mapNodeLabels.containsKey(lab))
                    {
                        mapNodeLabels.put(lab, setLabels.size());
                        setLabels.add(lab);
                    }
                    nodeLabs.add(mapNodeLabels.get(lab));
                }
                g.addNode(i,nodeLabs);
            }

            //Read edges
            br.readLine();
            while((str=br.readLine())!=null)
            {
                String[] split=str.split("\t");
                int source=Integer.parseInt(split[0]);
                int dest=Integer.parseInt(split[1]);
                String[] split2=split[2].split(",");
                TIntHashSet edgeLabs=new TIntHashSet();
                for(i=0;i<split2.length;i++)
                {
                    if(!mapEdgeLabels.containsKey(split2[i]))
                    {
                        mapEdgeLabels.put(split2[i],setLabels.size());
                        setLabels.add(split2[i]);
                    }
                    edgeLabs.add(mapEdgeLabels.get(split2[i]));
                }
                g.addEdge(source,dest,edgeLabs);
                if(edgeLabs.size()>maxMultiplicity)
                    maxMultiplicity=edgeLabs.size();
            }
            br.close();
        }
        catch(Exception e){
            System.out.println(e.getMessage());
        }
        return g;
    }

    public TCustomHashSet<Graph> readQueries(String queriesFile)
    {
        TCustomHashSet<Graph> setQueries=new TCustomHashSet<>(new GraphStrategy());
        try
        {
            BufferedReader br=new BufferedReader(new FileReader(queriesFile));
            String str=br.readLine();
            while(str!=null)
            {
                int numNodes=Integer.parseInt(br.readLine());
                boolean directed=false;
                if(queriesFile.endsWith("ged"))
                    directed=true;
                Graph q=new Graph(directed,numNodes);
                int i, j;
                //Read node labels
                for(i=0;i<numNodes;i++)
                {
                    String[] split=br.readLine().split(",");
                    TIntHashSet nodeLabs=new TIntHashSet();
                    for(j=0;j<split.length;j++)
                    {
                        String lab=split[j];
                        if(!mapNodeLabels.containsKey(lab))
                        {
                            mapNodeLabels.put(lab, setLabels.size());
                            setLabels.add(lab);
                        }
                        nodeLabs.add(mapNodeLabels.get(lab));
                    }
                    q.addNode(i,nodeLabs);
                }

                //Read edges
                br.readLine();
                while((str=br.readLine())!=null && !str.startsWith("#"))
                {
                    String[] split=str.split("\t");
                    int source=Integer.parseInt(split[0]);
                    int dest=Integer.parseInt(split[1]);
                    String[] split2=split[2].split(",");
                    TIntHashSet edgeLabs=new TIntHashSet();
                    for(i=0;i<split2.length;i++)
                    {
                        if(!mapEdgeLabels.containsKey(split2[i]))
                        {
                            mapEdgeLabels.put(split2[i],setLabels.size());
                            setLabels.add(split2[i]);
                        }
                        edgeLabs.add(mapEdgeLabels.get(split2[i]));
                    }
                    q.addEdge(source,dest,edgeLabs);
                }
                setQueries.add(q);
            }
            br.close();
        }
        catch(Exception e){
            System.out.println(e.toString());
        }
        return setQueries;
    }

    public void printQuery(Graph query)
    {
        System.out.print("QUERY GRAPH:\t");
        TIntHashSet[] nodeLabels=query.getNodeLabels();
        TIntObjectHashMap[] adjLists=query.getOutAdjList();
        int i;
        for(i=0;i<nodeLabels.length-1;i++)
        {
            TIntIterator it=nodeLabels[i].iterator();
            System.out.print(setLabels.get(it.next()));
            while(it.hasNext())
            {
                int labNode=it.next();
                System.out.print(","+setLabels.get(labNode));
            }
            System.out.print("---");
        }
        TIntIterator it=nodeLabels[i].iterator();
        System.out.print(setLabels.get(it.next()));
        while(it.hasNext())
        {
            int labNode=it.next();
            System.out.print(","+setLabels.get(labNode));
        }
        System.out.print("\t");
        for(i=0;i<adjLists.length;i++)
        {
            TIntIterator itEdge=adjLists[i].keySet().iterator();
            while(itEdge.hasNext())
            {
                int j=itEdge.next();
                if(query.isDirected() || i<j)
                {
                    System.out.print("("+i+","+j);
                    TIntHashSet edge=(TIntHashSet) adjLists[i].get(j);
                    TIntIterator itEdgeLab=edge.iterator();
                    while(itEdgeLab.hasNext())
                    {
                        String edgeLabel=setLabels.get(itEdgeLab.next());
                        System.out.print(","+edgeLabel);
                    }
                    System.out.print(")");
                }
            }
        }
        System.out.println();
    }

    public void writeQuery(Graph query)
    {
        try
        {
            bwDump.write("--------------------------------\n");
            bwDump.write("QUERY GRAPH:\t");
            TIntHashSet[] nodeLabels=query.getNodeLabels();
            TIntObjectHashMap[] adjLists=query.getOutAdjList();
            int i;
            for(i=0;i<nodeLabels.length-1;i++)
            {
                TIntIterator it=nodeLabels[i].iterator();
                bwDump.write(setLabels.get(it.next()));
                while(it.hasNext())
                    bwDump.write(","+setLabels.get(it.next()));
                bwDump.write("---");
            }
            TIntIterator it=nodeLabels[i].iterator();
            bwDump.write(setLabels.get(it.next()));
            while(it.hasNext())
                bwDump.write(","+setLabels.get(it.next()));
            bwDump.write("\t");
            for(i=0;i<adjLists.length;i++)
            {
                TIntIterator itEdge=adjLists[i].keySet().iterator();
                while(itEdge.hasNext())
                {
                    int j=itEdge.next();
                    if(query.isDirected() || i<j)
                    {
                        bwDump.write("("+i+","+j);
                        TIntHashSet edge=(TIntHashSet) adjLists[i].get(j);
                        TIntIterator itEdgeLab=edge.iterator();
                        while(itEdgeLab.hasNext())
                        {
                            String edgeLabel=setLabels.get(itEdgeLab.next());
                            bwDump.write(","+edgeLabel);
                        }
                        bwDump.write(")");
                    }
                }
            }
            bwDump.write("\n");
        }
        catch(Exception e)
        {
            System.out.println(e.getMessage());
        }
    }

    public void writeEmbedding(int[] solution, int[] map_node_to_state)
    {
        try
        {
            bwDump.write("[");
            int i;
            for(i=0;i<solution.length-1;i++)
                bwDump.write(solution[map_node_to_state[i]]+",");
            bwDump.write(solution[map_node_to_state[i]]+"]\n");
        }
        catch(Exception e)
        {
            System.out.println(e.getMessage());
        }
    }

    public void initDumpFile(String dumpFile)
    {
        try
        {
            bwDump=new BufferedWriter(new FileWriter(dumpFile));
            bwDump.close();
            bwDump=new BufferedWriter(new FileWriter(dumpFile,true));
        }
        catch(Exception e)
        {
            System.out.println(e.getMessage());
        }
    }

    public void closeDumpFile()
    {
        try
        {
            bwDump.close();
        }
        catch(Exception e)
        {
            System.out.println(e.getMessage());
        }
    }

}
