package matching;

import data.Graph;
import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.hash.TIntObjectHashMap;
import gnu.trove.set.hash.TIntHashSet;
import io.FileManager;

import java.util.BitSet;
import java.util.Vector;

public class RISolver
{
	private MatchingMachine mama;
	private Graph targetGraph;
	private String dumpFile;
	private long numOccurrences;
	private FileManager fm;
	
	public RISolver(Graph targetGraph, String dumpFile, FileManager fm)
	{
		this.targetGraph=targetGraph;
		this.dumpFile=dumpFile;
		this.fm=fm;
	}
	
	public void match(Graph queryGraph)
	{
		numOccurrences=0;

		BitSet[] domains=computeDomains(queryGraph);

		this.mama=new MatchingMachine(queryGraph);

		Vector<Integer>[] symmCond=queryGraph.getSymmetryConditions();

		TIntObjectHashMap[] targetOutAdjLists=targetGraph.getOutAdjList();
		TIntObjectHashMap[] targetInAdjLists=targetGraph.getInAdjList();
		int nofTargetNodes=targetOutAdjLists.length;
		int i, j;
		int nof_sn=mama.nof_sn;
		int[] parent_state=mama.parent_state;
		MamaParentType[] parent_type=mama.parent_type;
		int[] candidatesIT=new int[nof_sn];
		int[][] candidates = new int[nof_sn][];
		int[] solution = new int[nof_sn];
		for(i=0;i<nof_sn;i++)
			solution[i] = -1;
		TIntHashSet[] cmatched=new TIntHashSet[nof_sn];
		for(i=0;i<cmatched.length;i++)
			cmatched[i]=new TIntHashSet();
		boolean[] matched = new boolean[nofTargetNodes];

		for(i=0; i<nof_sn; i++)
		{
			if(parent_type[i] == MamaParentType.PARENTTYPE_NULL)
			{
				int n = mama.map_state_to_node[i];
				candidates[i]=new int[domains[n].cardinality()];
				int k = 0;
				for(j = domains[n].nextSetBit(0); j >= 0; j = domains[n].nextSetBit(j+1))
				{
					candidates[i][k]=j;
					k++;
				}
				candidatesIT[i]=-1;
			}
		}

		int psi = -1;
		int si = 0;
		int ci;
		int sip1;
		while(si != -1)
		{
			if(psi >= si)
				matched[solution[si]] = false;
			ci = -1;
			candidatesIT[si]++;
			while(candidatesIT[si]<candidates[si].length)
			{
				ci = candidates[si][candidatesIT[si]];
				solution[si] = ci;
				if(!matched[ci]
						&& domains[mama.map_state_to_node[si]].get(ci)
						&& condCheck(si,solution,symmCond)
						&& edgesCheck(si,solution)
					)
					break;
				else
					ci=-1;
				candidatesIT[si]++;
			}
			if(ci == -1)
			{
				psi = si;
				si--;
			}
			else
			{
				if(si == nof_sn -1)
				{
					numOccurrences++;
					if(dumpFile!=null)
						fm.writeEmbedding(solution,mama.map_node_to_state);
					psi = si;
				}
				else
				{
					matched[solution[si]] = true;
					sip1 = si+1;
					if(parent_type[sip1] != MamaParentType.PARENTTYPE_NULL)
					{
						if(parent_type[sip1] == MamaParentType.PARENTTYPE_IN)
							candidates[sip1] = targetInAdjLists[solution[parent_state[sip1]]].keys();
						else
							candidates[sip1] = targetOutAdjLists[solution[parent_state[sip1]]].keys();
					}
					candidatesIT[si+1]=-1;
					psi = si;
					si++;
				}
			}
		}
	}

	private BitSet[] computeDomains(Graph queryGraph)
	{
		int numQueryNodes=queryGraph.getNumNodes();
		int numTargetNodes=targetGraph.getNumNodes();
		BitSet[] domains=new BitSet[numQueryNodes];
		int i, j;
		for(i=0;i<domains.length;i++)
			domains[i]=new BitSet(numTargetNodes);

		TIntObjectHashMap[] targetOutAdjLists=targetGraph.getOutAdjList();
		TIntObjectHashMap[] targetInAdjLists=targetGraph.getInAdjList();
		TIntObjectHashMap[] queryOutAdjLists=queryGraph.getOutAdjList();
		TIntObjectHashMap[] queryInAdjLists=queryGraph.getInAdjList();
		TIntHashSet[] targetNodeLabs=targetGraph.getNodeLabels();
		TIntHashSet[] queryNodeLabs=queryGraph.getNodeLabels();
		for(i=0;i<numTargetNodes;i++)
		{
			//Find compatible query nodes and update domains
			for(j=0;j<domains.length;j++)
			{
				if(targetNodeLabs[i].containsAll(queryNodeLabs[j])
						&& queryOutAdjLists[j].size()<=targetOutAdjLists[i].size()
						&& queryInAdjLists[j].size()<=targetInAdjLists[i].size())
				{
					domains[j].set(i);
					//System.out.println(j+"-"+i);
				}
			}
		}

		int ra, qb, rb;
		boolean notfound;
		for(int qa=0; qa<queryNodeLabs.length; qa++)
		{
			for(ra = domains[qa].nextSetBit(0); ra >= 0; ra = domains[qa].nextSetBit(ra+1))
			{
				//for each edge qa->qb  check if exists ra->rb
				TIntIterator it=queryOutAdjLists[qa].keySet().iterator();
				while(it.hasNext())
				{
					qb = it.next();
					notfound = true;
					TIntHashSet setColsQuery = (TIntHashSet) queryOutAdjLists[qa].get(qb);
					TIntIterator it2=targetOutAdjLists[ra].keySet().iterator();
					while(it2.hasNext())
					{
						rb = it2.next();
						TIntHashSet setColsTarget = (TIntHashSet) targetOutAdjLists[ra].get(rb);
						if(domains[qb].get(rb) && setColsTarget.containsAll(setColsQuery)){
							notfound = false;
							break;
						}
					}
					if(notfound)
						domains[qa].set(ra, false);
				}
			}
		}

		return domains;
	}

	private boolean condCheck(int si, int[] solution, Vector<Integer>[] symmCond)
	{
		boolean condCheck=true;
		Vector<Integer> condNode=symmCond[mama.map_state_to_node[si]];
		int ii;
		for(ii=0;ii<condNode.size();ii++)
		{
			int targetNode=solution[mama.map_node_to_state[condNode.get(ii)]];
			if(solution[si]<targetNode)
			{
				condCheck = false;
				break;
			}
		}
		return condCheck;
	}

	private boolean edgesCheck(int si, int[] solution)
	{
		TIntObjectHashMap[] targetOutAdjLists=targetGraph.getOutAdjList();
		for(int me=0; me<mama.edges_sizes[si]; me++)
		{
			int source = solution[mama.edges[si][me].source];
			int target = solution[mama.edges[si][me].target];
			if(!targetOutAdjLists[source].containsKey(target))
				return false;
			else
			{
				TIntHashSet setColsQuery=mama.edges[si][me].attr;
				if(setColsQuery!=null)
				{
					TIntHashSet setColsTarget = (TIntHashSet) targetOutAdjLists[source].get(target);
					if(!setColsQuery.isEmpty() && !setColsTarget.containsAll(setColsQuery))
						return false;
				}
			}
		}
		return true;
	}

	public long getNumOccurrences()
	{
		return numOccurrences;
	}
}