package test;

import data.Graph;
import gnu.trove.set.hash.TCustomHashSet;
import io.FileManager;
import matching.RISolver;

public class TestRIMulti
{
    public static void main(String[] args)
    {
        //Input parameters
        String netFile="Networks/openflights.geu";
        String queryFile="Queries/openflights/4/openflights_4.geu";
        String dumpFile=null;
        //String dumpFile="dumpToy.txt";

        long inizio=System.currentTimeMillis();

        //Read input network
        System.out.println("Reading graph file...");
        FileManager fm=new FileManager();
        if(!netFile.endsWith("geu") && !netFile.endsWith("ged"))
        {
            System.out.println("Error! Unrecognizable file extension: "+netFile);
            System.exit(1);
        }
        Graph net=fm.readGraph(netFile);
        //System.out.println(net);

        //Read queries
        System.out.println("Reading queries file...");
        if(!queryFile.endsWith("geu") && !queryFile.endsWith("ged"))
        {
            System.out.println("Error! Unrecognizable file extension: "+queryFile);
            System.exit(1);
        }
        TCustomHashSet<Graph> setQueries=fm.readQueries(queryFile);

        //Perform subgraph matching
        System.out.println("Performing subgraph matching...\n");
        RISolver solver=new RISolver(net,dumpFile,fm);
        if(dumpFile!=null)
            fm.initDumpFile(dumpFile);
        for(Graph query : setQueries)
        {
            fm.printQuery(query);
            if(dumpFile!=null)
                fm.writeQuery(query);
            solver.match(query);
            long numOccs=solver.getNumOccurrences();
            System.out.println("Number of matches found: "+numOccs+"\n");
        }
        if(dumpFile!=null)
            fm.closeDumpFile();

        double fine=System.currentTimeMillis();
        double totalTime=(fine-inizio)/1000;

        System.out.println("Time elapsed: "+totalTime+" secs");
    }
}
